# -*- coding: windows-1250 -*-

# Program wywietlajcy struktur podanego pliku .cas i zapisujcy do drugiego pliku dane w nim zawarte
# Autor: Adam Salamon
# Data: 25-11-2009

import getopt
import sys
import string
import types

NAGL_ROZB_NORMAL = [0x55, 0x55]
NAGL_ROZB_T2600_1 = [0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xfe, 0xfc, 0xfc, 0xf8, 0xf0, 0xc7]
NAGL_ROZB_T2600_2 = [0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, 0xfc, 0xfc, 0xf8, 0xf0, 0xc7]

REK_PELNY = 0xfc
REK_NIEPELNY = 0xfa
REK_KONCOWY = 0xfe
DEFAULT_INIT_T2600 = 0xc2aa

def myhex(x, dlg):
	return hex(int(x))[2:].zfill(dlg)

class NotEnoughData(Exception):
	def __init__(self, value):
		self.value = value

class ExceptionEOF(Exception):
	pass

class BlednaSkladnia(Exception):
	pass

class BadT2600Block(Exception):
	pass

class MyError(Exception):
     def __init__(self, value):
         self.value = value

def myread(plik, dlg_req):
	s = plik.read(dlg_req)
	dlg = len(s)

	if dlg == 0: raise ExceptionEOF()
	if len(s) != dlg_req: raise MyError('Uszkodzony plik wejciowy.')

	return s

def konwertuj_litery_pl(s, tryb):
	listy = ('ʌӣѯ',
	          chr(164) + chr(168) + chr(151) + chr(143) + chr(224) + chr(157) +
	          chr(227) + chr(189) + chr(141) + chr(165) + chr(169) + chr(152) +
	          chr(134) + chr(162) + chr(136) + chr(228) + chr(190) + chr(171))

	s1 = ''
	for i in range(len(s)):
		nr = string.find(listy[tryb], s[i])
		if nr == -1:
			s1 = s1 + s[i]
		else:
			s1 = s1 + listy[1 - tryb][nr]
	return s1

def print_pol(s):
	print konwertuj_litery_pl(s, 0)

def czytaj_liczbe(plik_we, dlg):
	liczba = 0
	s = myread(plik_we, dlg)
	for i in range(dlg - 1, -1, -1):
		liczba = liczba * 256
		liczba += ord(s[i])
	return liczba

def begins_with(bajty, nagl):
	for i in range(len(nagl)):
		try:
			if bajty[i] != nagl[i]: return False
		except IndexError: # za mao danych na licie bajty
			return False

	return True

def typ_bloku_standard(typ, dlg, rodzaj):
	if rodzaj == REK_PELNY: return 'REKORD PENY'
	if rodzaj == REK_NIEPELNY: return 'REKORD NIEPENY'
	if rodzaj == REK_KONCOWY: return 'REKORD KOCOWY'
	return 'NIESTANDARDOWY'

def opis_rekordu(typ, bajty):
	dlg = len(bajty)

	if typ == 0:
		if dlg != 132: return 'NIESTANDARDOWA DG', None
		rodzaj = bajty[2]
		opis = typ_bloku_standard(typ, len(bajty), rodzaj)
		return opis, rodzaj

	elif typ == 1:
		return 'BLOK T2600', None

	elif typ == 2:
		if dlg != 142: return 'INNY T2600', None
		rodzaj = bajty[12]
		opis = typ_bloku_standard(typ, len(bajty), rodzaj)
		return opis + ' T2600', rodzaj

	else:
		return 'NIEROZPOZNANY',  None

def typ_rekordu(bajty):
	rek_dlg = len(bajty)

	dlg = len(bajty)
	if begins_with(bajty, NAGL_ROZB_NORMAL) and dlg >= 5:
		typ = 0

	elif begins_with(bajty, NAGL_ROZB_T2600_1):
		typ = 1

	elif begins_with(bajty, NAGL_ROZB_T2600_2):
		typ = 2

	else:
		typ = -1 # nierozpoznany

	opis, rodzaj = opis_rekordu(typ, bajty)
	return typ, opis, rodzaj

def oblicz_sume_kontrolna(bajty):
	suma = 0

	for i in range(len(bajty)):
		suma += bajty[i]
		if suma > 255: suma -= 255

	return suma
	
class NaglowekRozb():
	pass

class Segment_T2600():
	def __init__(self, pocz, kon, init, dane, status):
		self.pocz = pocz
		self.kon = kon
		self.init = init
		self.dane = dane
		self.status = status

class Blok_T2600:
	def __init__(self, bajty):
		self.segmenty = []
		self.czytaj_segmenty(bajty)

	def czytaj_segmenty(self, bajty):

		def room2(ile):
			if not (i + ile - 1 < dlg): raise BadT2600Block()

		i = 0
		dlg = len(bajty)
		dlg_rozbieg = len(NAGL_ROZB_T2600_1)

		while i < dlg:
			pocz_segm = i

			if begins_with(bajty[i:], NAGL_ROZB_T2600_1):
				i += dlg_rozbieg
				self.segmenty.append(NaglowekRozb())

			room2(9)
			if bajty[i + 7] != 0x65: raise BadT2600Block()

			pocz_ad = bajty[i] + (bajty[i + 1] << 8)
			kon_ad = bajty[i + 2] + (bajty[i + 3] << 8)
			init_ad = bajty[i + 4] + (bajty[i + 5] << 8)
			status = bajty[i + 6]

			if oblicz_sume_kontrolna(bajty[pocz_segm : i + 8]) != bajty[i + 8]: raise BadT2600Block()
			i += 9

			ile_danych = kon_ad - pocz_ad
			room2(ile_danych + 1)

			dane = bajty[i : i + ile_danych]
			if oblicz_sume_kontrolna(dane) != bajty[i + ile_danych]: raise BadT2600Block()

			seg = Segment_T2600(pocz_ad, kon_ad - 1, init_ad, dane, status)
			self.segmenty.append(seg)
			i += ile_danych + 1

class Plik_Wyjscowy:
	def __init__(self, nazwa_pliku):
		self.plik = file(nazwa_pliku, 'wb')
		self.dane = []
		self.first_record = True
		self.after_bin = False
		self.last_kon = None
		self.last_kon_index = None

	def __del__(self):
		if hasattr(self, 'plik'):
			for bajt in self.dane: self.plik.write(chr(bajt))
			self.plik.close()

	def write(self, d):
		typ = type(d)

		if typ == types.ListType:
			self.dane.extend(d)

		elif typ == types.StringType:
			for x in d: self.dane.append(ord(x))

		self.first_record = False
		self.after_bin = False

	def write_bin(self, pocz, kon, dane):
		if self.first_record: self.write([255, 255]) # tutaj nastpi ustawienie after_bin na false
		self.first_record = False

		if not g_optimize or not (self.after_bin and self.last_kon + 1 == pocz):
			self.last_kon_index = len(self.dane) + 2
			self.write([pocz % 256, pocz / 256, kon % 256, kon / 256])
			self.write(dane)
		
		else: # optymalizacja, dopisanie do poprzedniego bloku
			self.dane[self.last_kon_index] = kon % 256
			self.dane[self.last_kon_index + 1] = kon / 256
			self.write(dane)
		
		self.after_bin = True
		self.last_kon = kon

def czytaj_rekord(plik_we, plik_wy):
	global g_nr_rek, g_zapis_stop

	def ustaw_pocz_kon(bajty, rodzaj, offset):
		pocz = offset

		if rodzaj == REK_PELNY:
			kon = offset + 127
		
		elif rodzaj == REK_NIEPELNY:
			kon = offset + bajty[offset + 127] - 1
		
		elif len(bajty) - 2 >= 0: # rekord niestandardowy
			kon = len(bajty) - 2
		
		else:
			kon = len(bajty) - 1

		return pocz, kon

	rek_typ = myread(plik_we, 4)

	rek_dlg = czytaj_liczbe(plik_we, 2)
	rek_aux = czytaj_liczbe(plik_we, 2)

	if rek_typ == 'FUJI':
		s = myread(plik_we, rek_dlg)
		print_pol('PROGRAM: ' + s)

	elif rek_typ == 'baud':
		print_pol('PREDKO: ' + `rek_aux` + ' bodw')

	elif rek_typ == 'data':
		g_nr_rek += 1
		rek_pilot = '%.1f' % (float(rek_aux) / 1000)

		s = myread(plik_we, rek_dlg)
		bajty = map(ord, s)
		typ, opis, rodzaj = typ_rekordu(bajty)
		koncowy_rek = rodzaj == REK_KONCOWY

		# Sprawdzenie sum kontrolnych rekordu

		status = 'OK'
		if typ == 0 or typ == 2:
			suma = oblicz_sume_kontrolna(bajty[:-1])
			if suma != bajty[-1]: status = 'NOK'

		elif typ == 1:
			try:
				blok2600 = Blok_T2600(bajty)
			except BadT2600Block:
				status = 'NOK'
				blok2600 = None
		else:
			status = '???'

		zapis_txt = ''
		if g_zapis and not g_zapis_stop and g_nr_rek >= g_nr_rek_min and (g_nr_rek <= g_nr_rek_max or g_nr_rek_max == -1):
			if koncowy_rek and g_nr_rek_max == -1:
				g_zapis_stop = True
			else:
				zapis_txt = 'ZAPISUJ!'

		print_pol('REK: ' +  `g_nr_rek`.ljust(6) + 'PILOT: ' + (rek_pilot + 's').ljust(8) + 'DG: ' + `rek_dlg`.ljust(7) + \
		           opis.ljust(22) + status.ljust(6) + zapis_txt)

		# Wypisanie segmentw bloku Turbo 2600

		if g_verbose and typ == 1 and blok2600:
			print
			for seg in blok2600.segmenty:
				if isinstance(seg, Segment_T2600):
					s = '           Segment: $' + myhex(seg.pocz, 4) + ' - $' + myhex(seg.kon, 4) + ', status: $' + myhex(seg.status, 2)
					if seg.init != DEFAULT_INIT_T2600: s += ', init: $' + myhex(seg.init, 4)
					print_pol(s)
				else:
					print_pol('           Nagwek rozbiegowy')
			print

		# Zapis

		if not zapis_txt: return

		if not g_bytes_range:

			if typ == 1 and blok2600:
				for seg in blok2600.segmenty:
					if isinstance(seg, Segment_T2600):
						plik_wy.write_bin(seg.pocz, seg.kon, seg.dane)
						if seg.init != DEFAULT_INIT_T2600:
							plik_wy.write([0xe2, 0x02, 0xe3, 0x02, seg.init % 256, seg.init / 256])

			elif typ == 0:
				if len(bajty) == 132:
					offset = 3 # omijamy typ rekordu
				else:
					offset = 2 # za dwoma bajtami rozbiegowymi
				
				pocz, kon = ustaw_pocz_kon(bajty, rodzaj, offset)
				plik_wy.write(bajty[pocz : kon + 1])

			elif typ == 2:
				pocz, kon = ustaw_pocz_kon(bajty, rodzaj, 13)
				plik_wy.write(bajty[pocz : kon + 1])
				
		else: # podano zakres
			pocz = min(g_rek_byte_min, rek_dlg - 1)
			
			if g_rek_byte_max != -1:
				kon = min(g_rek_byte_max, rek_dlg - 1)
			else:
				kon = rek_dlg - 1

			plik_wy.write(bajty[pocz : kon + 1])

	else:
		raise MyError('Nieznany typ rekordu!')

def sprawdz_naglowek(plik_we):
	poz = plik_we.tell()
	s = myread(plik_we, 8)
	plik_we.seek(poz) # przywro jak byo
	if len(s) != 8 or s[:4] != 'FUJI': raise MyError('Plik wejciowy nie jest w formacie .cas')

def parsuj_zakres(s):
	x = string.find(s, '-')
	if x == -1: raise BlednaSkladnia()
	if s[:x] == '': raise BlednaSkladnia()

	try:
		if s[x + 1:] == '':
			max_value = -1
		else:
			max_value = int(s[x + 1:])
		return int(s[:x]), max_value

	except ValueError:
		raise BlednaSkladnia()

def usage():
	print_pol('\nSkadnia wywoania:\n')
	print_pol('python readcas.py [ -v ] [ -o ] plik_we [ [ [ plik_wy ] zakres_rek_do_zapisu ]\n       zakres_bajt_w_rek_do_zapisu ]\n')
	print_pol('-v - wywietla segmenty w blokach Turbo 2600')
	print_pol('-o - optymalizuje zapis blokw binarnych\n')
	print_pol('Zakresy podawa naley w postaci \'min-max\', np. 10-20.')
	print_pol('Rekordy numerowane s od 1, bajty w rekordzie od 0.\n')
	print_pol('Zapis \'wart-\' dla zakresu rekordw oznacza zapis do koca pliku albo do\nnajbliszego rekordu kocowego.\n')
	print_pol('Zapis \'wart-\' dla zakresu bajtw oznacza zapis do ostatniego bajtu w rekordzie.\n')
	print_pol('Brak podanego zakresu rekordw oznacza zapis od pierwszego rekordu.\n')
	print_pol('Brak podanego zakresu bajtw oznacza zapis zgodnie z typem rekordu: peny,      niepeny, niestandardowy, blok Turbo 2600, rekord 2600.')
	print_pol('Bloki Turbo 2600 przy braku zakresu bajtw zapisywane s binarnie.')
	print_pol('Rekordy niestandardowe (zaczynajce si bajtami $55, $55 o dugoci rnej od')
	print_pol('132) przy braku zakresu bajtw zapisywane s od trzeciego bajtu.')

def parse_commandline():
	global g_nazwa_pliku_we, g_nazwa_pliku_wy, g_zapis, g_bytes_range, g_nr_rek_min, g_nr_rek_max, g_rek_byte_min, g_rek_byte_max, \
	        g_verbose, g_optimize

	g_zapis = False
	g_bytes_range = False
	g_nr_rek_min = 1
	g_nr_rek_max = -1
	g_verbose = False
	g_optimize = False

	try:
		opts, args = getopt.getopt(sys.argv[1:], 'vo')

		dlg = len(args)
		if dlg == 0 or dlg > 4: raise BlednaSkladnia()

		g_nazwa_pliku_we = args[0]
		g_nazwa_pliku_wy = None

		if dlg >= 2:
			g_nazwa_pliku_wy = args[1]
			g_zapis = True

		if dlg > 2:
			g_nr_rek_min, g_nr_rek_max = parsuj_zakres(args[2])
			if g_nr_rek_min: g_nr_rek_min = max(g_nr_rek_min, 1)

		if dlg > 3:
			g_rek_byte_min, g_rek_byte_max = parsuj_zakres(args[3])
			if g_rek_byte_min: g_rek_byte_min = max(g_rek_byte_min, 1)
			g_bytes_range = True

		for o, a in opts:
			try:
				if o == '-v':
					g_verbose = True
				elif o == '-o':
					g_optimize = True
				else:
					raise BlednaSkladnia()
			except ValueError:
				raise BlednaSkladnia()

	except getopt.GetoptError:
		raise BlednaSkladnia()

def main():
	global g_nr_rek, g_zapis_stop

	g_zapis_stop = False
	g_nr_rek = 0

	try:
		parse_commandline()
		plik_we = file(g_nazwa_pliku_we, 'rb')
		sprawdz_naglowek(plik_we)

		plik_wy = None
		if g_zapis: plik_wy = Plik_Wyjscowy(g_nazwa_pliku_wy)

		while True:	czytaj_rekord(plik_we, plik_wy)

		plik_we.close()

	except IOError:
		print_pol('Wystpi bd przy operacji wejcia-wyjcia!')

	except BlednaSkladnia:
		usage()
		return

	except MyError, e:
		print_pol(e.value)
		return

	except ExceptionEOF:
		pass

main()