# -*- coding: UTF-8 -*-

# Program wywietla zawarto pliku *.com i zapisujcy do drugiego pliku wybrane bloki
# Autor: Adam Salamon
# Data: 28-04-2006

import sys
import string
import types

class NotEnoughData(Exception):
	def __init__(self, value):
		self.value = value

class BadHex(Exception):
	pass

def myread(plik, dlg):
	s = plik.read(dlg)
	if len(s) != dlg: raise NotEnoughData(len(s))
	return s

def konwertuj_litery_pl(s, tryb):
	listy = ("ʌӣѯ", "㽍侫")

	s1 = ''
	for i in range(len(s)):
		nr = string.find(listy[tryb], s[i])
		if nr == -1:
			s1 = s1 + s[i]
		else:
			s1 = s1 + listy[1 - tryb][nr]
	return s1

def czytaj_liczbe_z_napisu(s, dlg):
	liczba = 0
	for i in range(dlg - 1, -1, -1):
		liczba = liczba * 256
		liczba += ord(s[i])
	return liczba

def czytaj_liczbe(plik_we, dlg):
	s = myread(plik_we, dlg)
	return czytaj_liczbe_z_napisu(s, dlg)

def zapisz_liczbe(plik_wy, liczba, dlg):
	for i in range(dlg):
		plik_wy.write(chr(liczba % 256))
		liczba = liczba / 256

def myhex(x, dlg):
	return hex(x)[2:].zfill(dlg)

def pobierz_zakresy():
	if len(sys.argv) < 4: return []

	wynik = []
	for arg in sys.argv[3 : len(sys.argv) + 1]:
		x = string.find(arg, '-')
		if x == -1:
			wynik.append(int(arg))
		else:
			pocz = 1
			kon = -1

			if arg[:x] != '': pocz = int(arg[:x])
			if arg[x + 1:] != '': kon = int(arg[x + 1:])
			wynik.append([pocz, kon])

	return wynik

def czytaj_com(plik_we, plik_wy, zakresy):
	nr_blok = 0
	first_write = True

	while True:
		try:
			pocz = czytaj_liczbe(plik_we, 2)
		except NotEnoughData, e:
			if e.value == 0:
				return
			else:
				raise

		if pocz == 65535: continue
		kon = czytaj_liczbe(plik_we, 2)
		nr_blok = nr_blok + 1
		dane = myread(plik_we, kon - pocz + 1)

		init = ''
		run = ''
		s = 'BLOK: ' + `nr_blok`.ljust(5) + myhex(pocz, 4) + ' - ' + myhex(kon, 4) + '   '
		if pocz == 0x2e2 and kon == 0x2e3:
		   init = 'INIT: ' + myhex(czytaj_liczbe_z_napisu(dane, 2), 4) + '   '
		elif pocz == 0x2e0 and kon == 0x2e1:
			run = 'RUN: ' + myhex(czytaj_liczbe_z_napisu(dane, 2), 4)
		elif pocz == 0x2e0 and kon == 0x2e3:
			init = 'INIT: ' + myhex(czytaj_liczbe_z_napisu(dane[2:], 2), 4) + '   '
			run = 'RUN: ' + myhex(czytaj_liczbe_z_napisu(dane, 2), 4)
		s = s + (init + run).ljust(10 + 3 + 9)

		write = False;
		for x in zakresy:
			if type(x) == types.IntType:
				if nr_blok == x:
					write = True
					break
			elif type(x) == types.ListType:
				if nr_blok >= x[0] and (nr_blok <= x[1] or x[1] == -1):
					write = True
					break

		if write: s = s + '   ZAPISUJ!'

		print konwertuj_litery_pl(string.rstrip(s), 0)

		if write and plik_wy != None:
			if first_write: plik_wy.write(chr(0xff) + chr(0xff))
			first_write = False
			zapisz_liczbe(plik_wy, pocz, 2)
			zapisz_liczbe(plik_wy, kon, 2)
			plik_wy.write(dane)

def usage():
	print konwertuj_litery_pl('Skadnia: readcom.py plik_we [ plik_wy nr_blok_lub_zakres1 ... ]\n', 0)
	print konwertuj_litery_pl('gdzie:\n', 0)
	print konwertuj_litery_pl('nr_blok_lub_zakres jest liczb lub wyraeniem postaci <nr_pocz>-<nr_kon>\n(np. 1-5)\n', 0)
	print konwertuj_litery_pl('W przypadku niepodania numeru pocztkowego lub kocowego bloku przyjmowany jest', 0)
	print konwertuj_litery_pl('zakres od pierwszego bloku lub zakres do ostatniego.', 0)

def main():
	print
	if len(sys.argv) == 1 or len(sys.argv) == 3:
		usage()
		return

	try:
		zakresy = pobierz_zakresy()
	except ValueError:
		print konwertuj_litery_pl('Bd podczas odczytywania listy zakresw!\n', 0)
		usage()
		return

	try:
		plik_we = file(sys.argv[1], 'rb')
		plik_wy = None
		if len(sys.argv) > 2: plik_wy = file(sys.argv[2], 'wb')
	except IOError:
		print konwertuj_litery_pl('Bd podczas otwierania pliku!', 0)
		return

	try:
		czytaj_com(plik_we, plik_wy, zakresy)
		plik_we.close()
		if plik_wy != None: plik_wy.close()

	except NotEnoughData:
		print konwertuj_litery_pl('Niespodziewany koniec pliku.', 0)
		return
	except IOError:
		print konwertuj_litery_pl('Bd podczas odczytu pliku wejciowego!', 0)

main()
